/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.mcsla.core.metrics;

import eu.DECIDEh2020.appManager.models.Expression;
import eu.DECIDEh2020.appManager.models.Metric;
import eu.DECIDEh2020.appManager.models.ServiceObjective;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

public class ServiceObjectiveMetrics {

    private MetricsContext context;

    private ServiceObjective objective;

    private String csId;

    ServiceObjectiveMetrics(MetricsContext context, ServiceObjective objective) {
        this(context, null, objective);
    }

    ServiceObjectiveMetrics(MetricsContext context, String csId, ServiceObjective objective) {
        this.context = context;
        this.objective = objective;
        this.csId = csId;
    }

    public MetricsContext getContext() {
        return context;
    }

    public EvaluationResult evaluate(double measuredValue) {
        if (objective.getMetrics().isEmpty()) {
            return result(measuredValue, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
        }

        Metric metric = objective.getMetrics().get(0);
        Expression expression = metric.getExpression();
        if (expression.getExpressionLanguage().equals(Expressions.LANGUAGE_PREDEFINED)) {
            Predefined predefined = Predefined.valueOf(expression.getExpression());
            switch(predefined) {
                default:
                    return result(measuredValue, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
            }
        } else {
            return errorResult("Expression language '" + expression.getExpressionLanguage() + "' not yet supported.");
        }
    }

    public EvaluationResult<Double> evaluate(List<Long> measuredValues) {
        if (objective.getMetrics().isEmpty()) {
            return errorResult("No metrics defined.");
        }

        Metric metric = objective.getMetrics().get(0);
        return (EvaluationResult<Double>)evaluate(metric, measuredValues);
    }

    public EvaluationResult<Double> evaluate(long[] measuredValues) {
        return evaluate(Arrays.stream(measuredValues).boxed().collect(Collectors.toList()));
    }

    public EvaluationResult<Double> evaluate(double[] measuredValues) {
        if (objective.getMetrics().isEmpty()) {
            return errorResult("No metrics defined.");
        }
        Metric metric = objective.getMetrics().get(0);
        return (EvaluationResult<Double>)evaluate(metric, Arrays.stream(measuredValues).boxed().collect(Collectors.toList()));
    }

    public EvaluationResult<Double> evaluate(Map<String, List<Long>> measuredValues) {
        if (objective.getMetrics().isEmpty()) {
            return errorResult("No metrics defined.");
        }

        Metric metric = objective.getMetrics().get(0);
        List<Double> subValues = metric.getUnderlyingMetrics()
                .stream()
                .filter(m -> measuredValues.containsKey(m.getId()))
                .map(m -> (Double)evaluate(m, measuredValues.get(m.getId())).getMeasurementResultValue())
                .collect(Collectors.toList());

        if (subValues.isEmpty()) {
            return errorResult("No underlying metrics for the given keys found.");
        }

        return (EvaluationResult<Double>)evaluate(metric, subValues);
    }

    public EvaluationResult<Double> evaluate() {
        if (objective.getMetrics().isEmpty()) {
            return errorResult("No metrics defined.");
        }

        Metric metric = objective.getMetrics().get(0);
        List<EvaluationResult> results = context.restore(objective.getTermName());
        List<Double> subValues = results.stream().map(r -> (Double)r.getMeasurementResultValue()).collect(Collectors.toList());

        if (subValues.isEmpty()) {
            return errorResult("No stored metrics evaluation result found.");
        }

        return (EvaluationResult<Double>)evaluate(metric, subValues);
    }

    private <T extends Comparable<T>, V extends Comparable<V>> EvaluationResult<V> evaluate(Metric metric, List<T> values) {
        Expression expression = metric.getExpression();
        if (expression.getExpressionLanguage().equals(Expressions.LANGUAGE_PREDEFINED)) {
            Predefined predefined = Predefined.valueOf(expression.getExpression());
            switch(predefined) {
                case AVAILABILITY_AGGREGATION_SUMTYPE:
                    AvailabilityAggregationSumType availability = (AvailabilityAggregationSumType)Expressions.create(predefined);
                    List<Double> finalList = values.get(0) instanceof Long ? ((List<Long>)values).stream().map(Long::doubleValue).collect(Collectors.toList()) : (List<Double>)values;
                    Double value = availability.calculate(finalList, metric.getParameters());
                    return (EvaluationResult<V>)result(value, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
                case AVAILABILITY_MTBFMTTR:
                    AvailabilityMTBFMTTR mtbfmttr = (AvailabilityMTBFMTTR)Expressions.create(predefined);
                    Double value0 = mtbfmttr.calculate((List<Long>)values, metric.getParameters());
                    return (EvaluationResult<V>)result(value0, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
                case AVAILABILITY_UPTIME_BC:
                    AvailabilityUptimeBillingCycle uptimeBillingCycle = (AvailabilityUptimeBillingCycle)Expressions.create(predefined);
                    Double value1 = uptimeBillingCycle.calculate((List<Long>)values, metric.getParameters());
                    return (EvaluationResult<V>)result(value1, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
                case AGGREGATION_SUMTYPE:
                    AggregationSumType<Double> sumAggregation = (AggregationSumType<Double>)Expressions.create(predefined);
                    List<Double> finalList2 = values.get(0) instanceof Long ? ((List<Long>)values).stream().map(Long::doubleValue).collect(Collectors.toList()) : (List<Double>)values;
                    Double value2 = sumAggregation.calculate(finalList2, metric.getParameters());
                    return (EvaluationResult<V>)result(value2, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
                case AGGREGATION_MINTYPE:
                    AggregationMinType<Double> minAggregation = (AggregationMinType<Double>)Expressions.create(predefined);
                    List<Double> finalList3 = values.get(0) instanceof Long ? ((List<Long>)values).stream().map(Long::doubleValue).collect(Collectors.toList()) : (List<Double>)values;
                    Double value3 = minAggregation.calculate(finalList3, metric.getParameters());
                    return (EvaluationResult<V>)result(value3, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
                case AGGREGATION_MAXTYPE:
                    AggregationMaxType<Double> maxAggregation = (AggregationMaxType<Double>)Expressions.create(predefined);
                    List<Double> finalList4 = values.get(0) instanceof Long ? ((List<Long>)values).stream().map(Long::doubleValue).collect(Collectors.toList()) : (List<Double>)values;
                    Double value4 = maxAggregation.calculate(finalList4, metric.getParameters());
                    return (EvaluationResult<V>)result(value4, Statement.valueOf(objective.getConditionStatement()), Double.valueOf(objective.getValue()));
                case EMPTY:
                    return errorResult("Empty metric.");
                default:
                    return errorResult("Predefined metric '" + predefined + "' not found.");
            }
        } else {
            return errorResult("Expression language '" + expression.getExpressionLanguage() + "' not yet implemented.");
        }
    }

    private <V extends Comparable<V>> EvaluationResult<V> result(V result, Statement conditionStatement, V value) {
        EvaluationResult<V> res = new EvaluationResult<>();
        res.setTermName(objective.getTermName());
        res.setConditionStatement(conditionStatement, value);
        res.setMeasurementResultValue(result);
        if (csId != null) {
            context.store(csId, res);
        }
        return res;
    }

    private <T extends Comparable<T>> EvaluationResult<T> errorResult(String message) {
        EvaluationResult<T> res = new EvaluationResult<>();
        res.setTermName(objective.getTermName());
        res.setError(true);
        res.setErrorMessage(message);
        return res;
    }

}
