/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.mcsla.core.metrics;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.ImmutableBiMap;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Maps;
import eu.DECIDEh2020.appManager.AppDescriptionFactory;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;
import eu.DECIDEh2020.appManager.models.AppDescription;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.IOException;
import java.nio.file.FileSystems;
import java.nio.file.Path;
import java.util.*;

public class MetricsContextTest {

    private MetricsContext context;

    @BeforeClass
    public void init() {
        Path path = FileSystems.getDefault().getPath("src/test/resources/DECIDE.json");
        try {
            AppDescription appDescription = new ObjectMapper().readValue(path.toFile(), AppDescription.class);
            AppDescriptionFactory.validateAppDescription(appDescription);
            context = MetricsContext.create(appDescription);
        } catch (DECIDEValidationException | IOException e) {
            e.printStackTrace();
        }
    }

    @DataProvider(name = "csIdsDownTimes")
    public Object[][] createCsIdsDownTimes() {
        return new Object[][] {
                { "38", new long[] { 3000L, 4200L, 6300L } },
                { "15", new long[] { 5000L } },
                { "28", new long[] { 500L, 4400L, 3000L, 1000L } }
        };
    }

    @DataProvider(name = "aggregationValues")
    public Object[][] createAggregationValues() {
        return new Object[][] {
                { Predefined.AVAILABILITY_AGGREGATION_SUMTYPE, Arrays.asList(50.0, 50.0, 50.0), -50.0 },
                { Predefined.AVAILABILITY_AGGREGATION_SUMTYPE, Arrays.asList(99.9, 99.9, 99.9 ), 99.7 },
                { Predefined.AGGREGATION_SUMTYPE, Arrays.asList(100.00, 3000.00, 550.00 ), 3650.00 }
        };
    }

    @DataProvider(name = "microservicesMeasuredValues")
    public Object[][] createMicroservicesMeasuredValues() {
        return new Object[][] {
                {
                    ImmutableMap.of(
                        "9005a07c-1148-49bf-8c94-83421ce20e52", Arrays.asList(0L, 9000L, 10000L),
                        "3d16aa10-1ba0-43b2-b9c3-fa4f2c761688", Arrays.asList(0L, 30000L, 30100L),
                        "7a1837ea-2f5e-4fb2-aa57-a8261f2a5a83", Arrays.asList(0L, 100000L, 100250L))
                },
                {
                    ImmutableMap.of(
                        "9005a07c-1148-49bf-8c94-83421ce20e52", Arrays.asList(47L, 48L, 49L),
                        "3d16aa10-1ba0-43b2-b9c3-fa4f2c761688", Arrays.asList(99L, 100L,101L),
                        "7a1837ea-2f5e-4fb2-aa57-a8261f2a5a83", Arrays.asList(199L, 200L,201L))
                }
        };
    }

    @Test
    public void metricsContext() {
        try {
            MetricsContext context = MetricsContext.create(AppDescriptionFactory.createAppDescription("Test"));
            assert false;
        } catch (IllegalStateException e) {
            assert true : e.getMessage();
        }
    }

    @Test(dataProvider = "aggregationValues")
    public void metricsContextAggregate(Predefined predefined, List<Double> values, Double result) {
        Double aggregated = MetricsContext.aggregate(values, predefined);
        Double rounded = Math.round(aggregated * 100.0) / 100.0;

        assert rounded.equals(result) : "Aggregation is " + rounded;
    }

    // Aggregation with explicit values.
    @Test
    public void availabilityAggregationRaw() {

        // Get all service objectives from the MCSLA (App SLA)
        Map<String, ServiceObjectiveMetrics> metrics = context.getAppMetrics();

        // Get the service objective for Availability
        ServiceObjectiveMetrics availabilityMetrics = metrics.get("Availability");

        // Evaluate against a list of measured cloud service values
        double[] values = { 99.9, 99.99, 99.95, 99.95 };
        EvaluationResult<Double> evaluation = availabilityMetrics.evaluate(values);

        // Read out the evaluation result
        System.out.println("result of raw: " + evaluation.getMeasurementResultValue());
        System.out.println("final condition met: " + evaluation.isConditionMet());

        assert !evaluation.isError() : evaluation.getErrorMessage();
    }

    // Aggregation based on internal recall. Precondition is that all values to be aggregated are measured once before.
    @Test(dependsOnMethods = "availabilityBillingCycle")
    public void availabilityAggregationMemory() {

        // Get all service objectives from the MCSLA (App SLA)
        Map<String, ServiceObjectiveMetrics> metrics = context.getAppMetrics();

        // Get the service objective for Availability
        ServiceObjectiveMetrics availabilityMetrics = metrics.get("Availability");

        // Evaluate against the previously measured and internally stored cloud service values
        EvaluationResult<Double> evaluation = availabilityMetrics.evaluate();

        // Read out the evaluation result
        System.out.println("final result from memory: " + evaluation.getMeasurementResultValue());
        System.out.println("final condition met: " + evaluation.isConditionMet());

        assert !evaluation.isError() : evaluation.getErrorMessage();
    }

    @Test(dataProvider = "csIdsDownTimes")
    public void availabilityBillingCycle(String csId, long[] values) {

        // Get all service objectives for a specific cloud service
        Map<String, ServiceObjectiveMetrics> metrics = context.getCsMetrics(csId);

        // Get the service objective for Availability
        ServiceObjectiveMetrics availabilityMetrics = metrics.get("Availability");

        // Evaluate against the measured values
        EvaluationResult<Double> evaluation = availabilityMetrics.evaluate(values);

        // Read out the evaluation result
        System.out.println("result of billing cycle: " + evaluation.getMeasurementResultValue());
        System.out.println("condition met: " + evaluation.isConditionMet());

        assert !evaluation.isError() : evaluation.getErrorMessage();
    }

    @Test(dataProvider = "microservicesMeasuredValues")
    public void availabilityAggregation(Map<String, List<Long>> measuredValues) {
        // Get all service objectives from the MCSLA (App SLA)
        Map<String, ServiceObjectiveMetrics> metrics = context.getAppMetrics();

        // Get the service objective for Availability
        ServiceObjectiveMetrics availabilityMetrics = metrics.get("Availability");

        // Evaluate against a list of measured cloud service values
        EvaluationResult<Double> evaluation = availabilityMetrics.evaluate(measuredValues);

        // Read out the evaluation result
        System.out.println("result of microservice: " + evaluation.getMeasurementResultValue());
        System.out.println("final condition met: " + evaluation.isConditionMet());

        assert !evaluation.isError() : evaluation.getErrorMessage();
    }
}
