/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.mcsla.service;

import eu.DECIDEh2020.mcsla.service.repositories.RepositoriesService;
import eu.DECIDEh2020.mcsla.service.repositories.RepositoriesVerticle;
import io.vertx.config.ConfigRetriever;
import io.vertx.config.ConfigRetrieverOptions;
import io.vertx.config.ConfigStoreOptions;
import io.vertx.core.*;
import io.vertx.core.http.HttpMethod;
import io.vertx.core.http.HttpServer;
import io.vertx.core.http.HttpServerOptions;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.dropwizard.MetricsService;
import io.vertx.ext.healthchecks.HealthCheckHandler;
import io.vertx.ext.healthchecks.Status;
import io.vertx.ext.web.Router;
import io.vertx.ext.web.RoutingContext;
import io.vertx.ext.web.api.RequestParameter;
import io.vertx.ext.web.api.RequestParameters;
import io.vertx.ext.web.api.contract.RouterFactoryOptions;
import io.vertx.ext.web.api.contract.openapi3.OpenAPI3RouterFactory;
import io.vertx.ext.web.handler.CorsHandler;
import io.vertx.ext.web.handler.StaticHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class MainVerticle extends AbstractVerticle {

    private Logger log = LoggerFactory.getLogger(getClass());

    private RepositoriesService repositories;

    @Override
    public void start(Future<Void> startFuture) {
        MetricsService metrics = MetricsService.create(vertx);

        ConfigStoreOptions storeOptions = new ConfigStoreOptions()
                .setType("env")
                .setConfig(new JsonObject().put("keys", new JsonArray().add("DECIDE_REPOS_BASE_DIR").add("DECIDE_ACSMI_DISCOVERY_SERVICE_URI")));

        ConfigRetriever retriever = ConfigRetriever.create(vertx, new ConfigRetrieverOptions().addStore(storeOptions));
        Future<JsonObject> configFuture = ConfigRetriever.getConfigAsFuture(retriever);

        configFuture.compose(config -> {
            log.debug("config read");

            Future<Void> repositoriesFuture = Future.future();
            vertx.deployVerticle(RepositoriesVerticle.class, new DeploymentOptions().setWorker(true).setConfig(config), result -> {
                if (result.succeeded()) {
                    log.debug("repositories verticle deployed successfully");
                    repositories = RepositoriesService.createProxy(vertx, "decide.mcsla.service.repositories.queue");
                    repositoriesFuture.complete();
                } else {
                    repositoriesFuture.fail(result.cause());
                }
            });

            repositoriesFuture.setHandler(res -> {
                if (res.succeeded()) {
                    log.debug("all verticles deployed");

                    OpenAPI3RouterFactory.create(vertx, "webroot/openapi.yaml", ar -> {
                        if (ar.succeeded()) {
                            OpenAPI3RouterFactory routerFactory = ar.result();
                            RouterFactoryOptions roptions = new RouterFactoryOptions().setMountNotImplementedHandler(true).setMountValidationFailureHandler(true);
                            routerFactory.setOptions(roptions);
                            routerFactory.addHandlerByOperationId("initApplication", this::handleInitApplication);
                            routerFactory.addHandlerByOperationId("getApplication", this::handleGetApplication);
                            routerFactory.addHandlerByOperationId("resetApplication", this::handleResetApplication);
                            routerFactory.addHandlerByOperationId("aggregateServiceObjective", this::handleAggregation);
                            routerFactory.addHandlerByOperationId("setMcsla", this::handleMcsla);

                            Router router = routerFactory.getRouter();
                            router.route().order(0).handler(CorsHandler.create("*").allowedMethods(Stream.of(HttpMethod.GET, HttpMethod.POST).collect(Collectors.toSet())));

                            router.route("/*").handler(StaticHandler.create());

                            HealthCheckHandler hch = HealthCheckHandler.create(vertx);
                            hch.register("simple-ok", future -> future.complete(Status.OK()));
                            hch.register("info", future -> future.complete(Status.OK(new JsonObject().put("build", 1))));
                            router.get("/health").handler(hch);

                            router.get("/metrics").produces("application/json").handler(routingContext -> {
                                JsonObject m = metrics.getMetricsSnapshot(vertx);
                                if (m != null) {
                                    routingContext.response().setStatusCode(200).end(m.encodePrettily());
                                } else {
                                    routingContext.response().setStatusCode(503).end();
                                }
                            });

                            HttpServer server = vertx.createHttpServer(new HttpServerOptions().setPort(8080));
                            server.requestHandler(router::accept).listen();

                            startFuture.complete();
                        } else {
                            // Something went wrong during router factory initialization
                            startFuture.fail(ar.cause());
                        }
                    });
                } else {
                    startFuture.fail(res.cause());
                }
            });

        }, startFuture);
    }

    private void handleInitApplication(RoutingContext context) {
        RequestParameters params = context.get("parsedParameters");

        RequestParameter uriParam = params.queryParameter("uri");
        String uri = uriParam != null && uriParam.isString() && !uriParam.isEmpty() ? uriParam.getString() : "";

        RequestParameter tokenParam = params.queryParameter("token");
        String token = tokenParam != null && tokenParam.isString() && !tokenParam.isEmpty() ? tokenParam.getString() : "";

        RequestParameter usernameParam = params.queryParameter("username");
        String username = usernameParam != null && usernameParam.isString() && !usernameParam.isEmpty() ? usernameParam.getString() : "";

        RequestParameter passwordParam = params.queryParameter("password");
        String password = passwordParam != null && passwordParam.isString() && !passwordParam.isEmpty() ? passwordParam.getString() : "";

        RequestParameter pathParam = params.queryParameter("path");
        String path = pathParam != null && pathParam.isString() && !pathParam.isEmpty() ? pathParam.getString() : null;

        if (path != null) {

        } else {
            repositories.initRepository(uri, token, username, password, res -> {
                JsonObject result = null;
                if (res.succeeded()) {
                    result = res.result();
                    result.put("success", true);
                } else {
                    result = new JsonObject().put("success", false);
                    result.put("error", new JsonObject().put("message", res.cause().getMessage()));
                }
                context.response().end(result.toBuffer());
            });
        }
    }

    private void handleGetApplication(RoutingContext context) {
        RequestParameters params = context.get("parsedParameters");

        RequestParameter nameParam = params.pathParameter("name");
        String name = nameParam != null && nameParam.isString() && !nameParam.isEmpty() ? nameParam.getString() : "";

        repositories.getApplication(name, res -> {
            JsonObject result = new JsonObject();
            if (res.succeeded()) {
                result.put("success", true);
                result.put("app", res.result());
            } else {
                result.put("success", false);
                result.put("error", new JsonObject().put("message", res.cause().getMessage()));
            }
            context.response().end(result.toBuffer());
        });
    }

    private void handleResetApplication(RoutingContext context) {
        RequestParameters params = context.get("parsedParameters");

        RequestParameter nameParam = params.pathParameter("name");
        String name = nameParam != null && nameParam.isString() && !nameParam.isEmpty() ? nameParam.getString() : "";

        repositories.resetApplication(name, res -> {
            JsonObject result = new JsonObject();
            if (res.succeeded()) {
                result.put("success", true);
                result.put("app", res.result());
            } else {
                result.put("success", false);
                result.put("error", new JsonObject().put("message", res.cause().getMessage()));
            }
            context.response().end(result.toBuffer());
        });
    }

    private void handleAggregation(RoutingContext context) {
        RequestParameters params = context.get("parsedParameters");

        RequestParameter nameParam = params.pathParameter("name");
        String name = nameParam != null && nameParam.isString() && !nameParam.isEmpty() ? nameParam.getString() : "";

        RequestParameter termParam = params.pathParameter("termName");
        String term = termParam != null && termParam.isString() && !termParam.isEmpty() ? termParam.getString() : "";

        RequestParameter predefinedParam = params.queryParameter("predefined");
        String predefined = predefinedParam != null && predefinedParam.isString() && !predefinedParam.isEmpty() ? predefinedParam.getString() : "";

        repositories.aggregateServiceObjectives(name, term, predefined, res -> {
            JsonObject result = new JsonObject();
            if (res.succeeded()) {
                result.put("success", true);
                result.put("aggregated", res.result());
            } else {
                result.put("success", false);
                result.put("error", new JsonObject().put("message", res.cause().getMessage()));
            }
            context.response().end(result.toBuffer());
        });
    }

    private void handleMcsla(RoutingContext context) {
        RequestParameters params = context.get("parsedParameters");

        RequestParameter nameParam = params.pathParameter("name");
        String name = nameParam != null && nameParam.isString() && !nameParam.isEmpty() ? nameParam.getString() : "";

        JsonObject mcsla = context.getBodyAsJson();

        repositories.updateMcsla(name, mcsla, res -> {
            JsonObject result = new JsonObject();
            if (res.succeeded()) {
                result.put("success", true);
                result.put("mcsla", res.result());
            } else {
                result.put("success", false);
                result.put("error", new JsonObject().put("message", res.cause().getMessage()));
            }
            context.response().end(result.toBuffer());
        });

    }

    public static void main(String[] args) {
        String[] params = Arrays.copyOf(args, args.length + 1);
        params[params.length - 1] = MainVerticle.class.getName();
        Launcher.executeCommand("run", params);
    }

}
