/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.mcsla.service.repositories;

import eu.DECIDEh2020.appManager.models.*;
import eu.DECIDEh2020.mcsla.core.metrics.Expressions;
import eu.DECIDEh2020.mcsla.core.metrics.Predefined;
import eu.DECIDEh2020.mcsla.core.metrics.Statement;

import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;

public class SlaFactory {

    static SlaFactory getInstance() {
        return new SlaFactory();
    }

    public Sla createFromNfrs(List<Nfr> nfrs, List<Microservice> microservices) {
        Sla sla = new Sla();
        sla.getObjectives().addAll(nfrs.stream().map(nfr -> toServiceObjective(nfr, microservices)).collect(Collectors.toList()));
        return sla;
    }

    private ServiceObjective toServiceObjective(Nfr nfr, List<Microservice> microservices) {
        ServiceObjective objective = new ServiceObjective();

        objective.setTermName(nfr.getType());

        if (nfr instanceof AvailabilityNfr) {
            AvailabilityNfr availability = (AvailabilityNfr)nfr;
            objective.setType("slo");
            objective.setValue(String.valueOf(availability.getValue()));
            objective.setUnit(availability.getUnit());
            objective.setConditionStatement(Statement.greaterOrEqual.name());
            Metric metric = new Metric();
            metric.setId("CSA_AV_001");
            metric.setScale("ratio");
            metric.setDescriptor("Aggregation of microservices");
            Expression expression = new Expression();
            expression.setExpression(Predefined.AVAILABILITY_AGGREGATION_SUMTYPE.name());
            expression.setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
            expression.setNote(Predefined.AVAILABILITY_AGGREGATION_SUMTYPE.note());
            expression.setUnit(availability.getUnit());
            metric.setExpression(expression);

            for (Microservice microservice : microservices) {
                Metric msMetric = new Metric();
                msMetric.setId(microservice.getId());
                msMetric.setNote(microservice.getName());
                msMetric.setDescriptor("Availability based on meantime between failure and meantime between recovery");
                msMetric.setScale("ratio");
                msMetric.setExpression(new Expression());
                msMetric.getExpression().setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
                msMetric.getExpression().setExpression(Predefined.AVAILABILITY_MTBFMTTR.name());
                msMetric.getExpression().setUnit("percentage");
                msMetric.getExpression().setNote(Predefined.AVAILABILITY_MTBFMTTR.note());
                metric.getUnderlyingMetrics().add(msMetric);
            }
            objective.getMetrics().add(metric);
        } else if (nfr instanceof PerformanceNfr) {
            PerformanceNfr performance = (PerformanceNfr)nfr;
            objective.setType("slo");
            objective.setValue(String.valueOf(performance.getValue()));
            objective.setUnit(performance.getUnit());
            objective.setConditionStatement(Statement.lessOrEqual.name());
            Metric metric = new Metric();
            metric.setId("CSA_PE_001");
            metric.setDescriptor("Maximum response time for a request");
            metric.setScale("ratio");
            Expression expression = new Expression();
            expression.setExpression(Predefined.AGGREGATION_MAXTYPE.name());
            expression.setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
            expression.setNote(Predefined.AGGREGATION_MAXTYPE.note());
            expression.setUnit(performance.getUnit());
            metric.setExpression(expression);

            for (Microservice microservice : microservices) {
                Metric msMetric = new Metric();
                msMetric.setId(microservice.getId());
                msMetric.setNote(microservice.getName());
                msMetric.setScale("ratio");
                msMetric.setExpression(new Expression());
                msMetric.getExpression().setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
                msMetric.getExpression().setExpression(Predefined.EMPTY.name());
                msMetric.getExpression().setNote(Predefined.EMPTY.note());
                msMetric.getExpression().setUnit("millisecond");
                metric.getUnderlyingMetrics().add(msMetric);
            }
            objective.getMetrics().add(metric);
        } else if (nfr instanceof ScalabilityNfr) {
            ScalabilityNfr scalability = (ScalabilityNfr)nfr;
            objective.setType("sqo");
            objective.setValue(String.valueOf(scalability.getValue()));
            objective.setConditionStatement(Statement.equal.name());
        } else if (nfr instanceof LocationNfr) {
            LocationNfr location = (LocationNfr)nfr;
            objective.setType("sqo");
            objective.setValue(location.getValue().get(0));
            objective.setConditionStatement(Statement.equal.name());
        } else if (nfr instanceof CostNfr) {
            CostNfr cost = (CostNfr)nfr;
            objective.setType("slo");
            objective.setValue(String.valueOf(cost.getValue()));
            objective.setUnit(cost.getUnit());
            objective.setConditionStatement(Statement.less.name());
            Metric metric = new Metric();
            metric.setId("CSA_CO_001");
            metric.setDescriptor("Sum of all costs");
            metric.setScale("ratio");
            Expression expression = new Expression();
            expression.setExpression(Predefined.AGGREGATION_SUMTYPE.name());
            expression.setNote(Predefined.AGGREGATION_SUMTYPE.note());
            expression.setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
            expression.setUnit(cost.getUnit());
            metric.setExpression(expression);
            objective.getMetrics().add(metric);
        }

        return objective;
    }

}
