/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.mcsla.service.repositories;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import eu.DECIDEh2020.appManager.models.Expression;
import eu.DECIDEh2020.appManager.models.Metric;
import eu.DECIDEh2020.appManager.models.ServiceObjective;
import eu.DECIDEh2020.appManager.models.Sla;
import eu.DECIDEh2020.mcsla.core.metrics.Expressions;
import eu.DECIDEh2020.mcsla.core.metrics.Predefined;
import eu.DECIDEh2020.mcsla.core.metrics.Statement;
import io.vertx.core.json.Json;
import io.vertx.core.json.JsonArray;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.client.WebClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collector;
import java.util.stream.Collectors;

public class SlaImport {

    private static final Logger log = LoggerFactory.getLogger(SlaImport.class);

    static Map<String, Sla> fromResult(JsonArray result) {
        Map<String, Sla> slas = new HashMap<>();

        result.stream().filter(node -> ((JsonObject)node).getValue("nfrvalue") != null).forEach((node -> {
            JsonObject objectiveNode = (JsonObject)node;
            String serviceId = String.valueOf(objectiveNode.getInteger("serviceid"));
            Sla sla = slas.get(serviceId);
            if (!slas.containsKey(serviceId)) {
                sla = new Sla();
                sla.setDescription(objectiveNode.getString("servicename"));
                sla.getCoveredServices().add(objectiveNode.getString("serviceclassname"));
                slas.put(serviceId, sla);
            }

            ServiceObjective objective = new ServiceObjective();
            sla.getObjectives().add(objective);

            String term = objectiveNode.getString("nfrmetric");
            objective.setTermName(term);
            if (objectiveNode.containsKey("nfrname") && !objectiveNode.getString("nfrname").equals(objective.getTermName())) {
                objective.setComment(objectiveNode.getString("nfrname"));
            }
            objective.setValue(objectiveNode.getString("nfrvalue"));

            if ("Region".equalsIgnoreCase(term)) {
                objective.setType("sqo");
                objective.setConditionStatement(Statement.equal.name());

            } else if ("Zone".equalsIgnoreCase(term)) {
                objective.setType("sqo");
                objective.setConditionStatement(Statement.equal.name());

            } else if ("Provider".equalsIgnoreCase(term)) {
                objective.setType("sqo");
                objective.setConditionStatement(Statement.equal.name());

            } else if ("Availability".equalsIgnoreCase(term)) {
                objective.setType("slo");
                objective.setUnit("percentage");
                objective.setConditionStatement(Statement.greaterOrEqual.name());
                Metric metric = new Metric();
                metric.setId("CS_" + serviceId + "_001");
                metric.setDescriptor("Uptime during a specific time period");
                metric.setScale("ratio");
                Expression expression = new Expression();
                expression.setNote(Predefined.AVAILABILITY_UPTIME_BC.note());
                expression.setExpression(Predefined.AVAILABILITY_UPTIME_BC.name());
                expression.setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
                expression.setUnit("percentage");
                metric.setExpression(expression);
                objective.getMetrics().add(metric);

            } else if ("Performance".equalsIgnoreCase(term)) {
                objective.setType("slo");
                objective.setUnit("millisecond");
                objective.setConditionStatement(Statement.lessOrEqual.name());
                Metric metric = new Metric();
                metric.setId("CS_" + serviceId + "_001");
                metric.setDescriptor("Response time for a request");
                metric.setScale("interval");
                Expression expression = new Expression();
                expression.setNote("delta(t)");
                expression.setExpression(Predefined.EMPTY.name());
                expression.setExpressionLanguage(Expressions.LANGUAGE_PREDEFINED);
                expression.setUnit("millisecond");
                metric.setExpression(expression);
                objective.getMetrics().add(metric);

            } else if ("Legal Level".equalsIgnoreCase(term)) {
                objective.setType("sqo");
                objective.setConditionStatement(Statement.equal.name());

            } else if ("Cost/Currency".equalsIgnoreCase(term)) {
                objective.setTermName("Cost");
                objective.setType("slo");
                objective.setConditionStatement(Statement.less.name());
                objective.setUnit("euro");

            }
        }));

        return slas;
    }

    static Map<String, Sla> fromResource(List<String> csIds) {

        Map<String, Sla> slas = new HashMap<>();

        try {
            InputStream input = RepositoriesServiceImpl.class.getClassLoader().getResourceAsStream("acsmi-services.json");
            JsonNode node = new ObjectMapper().readTree(input);

            if (node.isArray()) {
                slas = fromResult(new JsonArray(Json.encode(node)));
            }
        } catch (IOException e) {
            log.error("fetching slas for cloud services", e);
        }
        return slas.entrySet().stream().filter(entry -> csIds.contains(entry.getKey())).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
    }

}
