import { all, call, select, put, takeLatest } from 'redux-saga/effects'

import {
    FETCH_DECIDE_APP_REQUEST,
    fetchDecideAppRequest,
    fetchDecideAppResponse,
    fetchDecideAppFailure,
    UPDATE_MCSLA_REQUEST,
    updateMcslaResponse,
    updateMcslaFailure,
    AGGREGATE_OBJECTIVES_REQUEST,
    aggregateObjectivesResponse,
    aggregateObjectivesFailure
} from '../actions/mcsla'

import { 
    FETCH_SERVICE_URI_REQUEST,
    fetchServiceUriResponse,
    fetchServiceUriFailure,
    RESET_REPOSITORY_REQUEST,
    resetRepositoryResponse,
    resetRepositoryFailure
} from '../actions/repository';

export default function* rootSaga() {
    yield all([
        watchFetchServiceUri(),
        watchFetchDecideApp(),
        watchUpdateMcsla(),
        watchAggregateObjectives(),
        watchResetRepository()
    ])
}

function* fetchServiceUri({ from }) {
    try {
        const data = yield call(fetch, from);
        const uri = yield call([data, data.json]);
        yield put(fetchServiceUriResponse(uri));
        yield put(fetchDecideAppRequest());
    } catch (error) {
        console.log(error);
        yield put(fetchServiceUriFailure(error))
    }
}

function* watchFetchServiceUri() {
    yield takeLatest(FETCH_SERVICE_URI_REQUEST, fetchServiceUri)
}

export const getRepository = state => state.repository;
export const getApp = state => state.app;

function* fetchDecideApp() {
    const repository = yield select(getRepository);
    if (!repository.uri) {
        console.log("No repo configured.");
        yield put(fetchDecideAppFailure({message: "No repo configured!"}))
    } else {
        const app = yield select(getApp);

        let uri = app.serviceUri + '/applications/init?uri=' + encodeURIComponent(repository.uri);
        uri += repository.token ? '&token=' + repository.token : '&username=' + repository.username + '&password=' + repository.password;

        console.log('fetching from ' + uri);
        try {
            const data = yield call(fetch, uri);
            const value = yield call([data, data.json]);
            if (value.success) {
                yield put(fetchDecideAppResponse(value.name, value.app))
            } else {
                yield put(fetchDecideAppFailure(value.error))
            }
        } catch (error) {
            console.log(error);
            yield put(fetchDecideAppFailure(error))
        }
    }
}

function* watchFetchDecideApp() {
    yield takeLatest(FETCH_DECIDE_APP_REQUEST, fetchDecideApp)
}

function* updateMcsla({ name, mcsla }) {
    const app = yield select(getApp);
    const uri = app.serviceUri + '/applications/' + name + '/mcsla';
    console.log('updating mcsla to ' + uri);

    try {
        const data = yield call(fetch, uri, {
            method: 'POST',
            body: JSON.stringify(mcsla)
        })
        const value = yield call([data, data.json]);
        if (value.success) {
            yield put(updateMcslaResponse(value))
        } else {
            yield put(updateMcslaFailure(value.error))
        }
    } catch (error) {
        console.log(error);
        yield put(updateMcslaFailure(error))
    }
}

function* watchUpdateMcsla() {
    yield takeLatest(UPDATE_MCSLA_REQUEST, updateMcsla)
}

function* aggregateObjectives({ termName, predefined }) {
    const app = yield select(getApp);
    const uri = app.serviceUri + '/applications/' + app.appName + '/mcsla/' + termName + '/aggregate?predefined=' + predefined;
    console.log('aggregate objectives: ' + uri);

    try {
        const data = yield call(fetch, uri);
        const value = yield call([data, data.json]);
        if (value.success) {
            yield put(aggregateObjectivesResponse(termName, predefined, value.aggregated))
        } else {
            yield put(aggregateObjectivesFailure(value.error))
        }
    } catch (error) {
        console.log(error);
        yield put(aggregateObjectivesFailure(error))
    }
}

function* watchAggregateObjectives() {
    yield takeLatest(AGGREGATE_OBJECTIVES_REQUEST, aggregateObjectives)
}

function* resetRepository() {
    const app = yield select(getApp);
    const uri = app.serviceUri + '/applications/' + app.appName + '/reset';
    console.log('reset repository: ' + uri);

    try {
        const data = yield call(fetch, uri);
        const value = yield call([data, data.json]);
        if (value.success) {
            yield put(resetRepositoryResponse(value.app))
        } else {
            yield put(resetRepositoryFailure(value.error))
        }
    } catch (error) {
        console.log(error);
        yield put(resetRepositoryFailure(error))
    }
}

function* watchResetRepository() {
    yield takeLatest(RESET_REPOSITORY_REQUEST, resetRepository)
}

